# -*- coding: utf-8 -*-

"""Module designed to the BOLD signal synthesis according to the Linear and
Time Invariant model described in:

  - Makni, S., Ciuciu, P., Idier, J., & Poline, J.-B. (2005). Joint
    detection-estimation of brain activity in functional MRI: a Multichannel
    Deconvolution solution. IEEE Transactions on Signal Processing, 53(9),
    3488–3502. http://doi.org/10.1109/TSP.2005.853303

It also provides a set of plotting functions (based on matplotlib) - see
L{bolsynth.plot}
"""
# import pyhrf.boldsynth.base
# import pyhrf.boldsynth.noise
# import pyhrf.boldsynth.spatialconfig 
# import pyhrf.boldsynth.hrf 
# import pyhrf.boldsynth.nrl 
# import pyhrf.boldsynth.drift 
# import pyhrf.boldsynth.boldmodel 
# import pyhrf.boldsynth.paradigm
#import plot

#import pyhrf.boldsynth.scenarios 
