# -*- coding: utf-8 -*-

import unittest

import numpy as np

import pyhrf

from core_test import *
from boldsynthTest import *
from toolsTest import *
from commandTest import *
from test_xml import *
from statsTest import *
from graphtest import *
from analysertest import *
from jdetest import *
from iotest import *
from test_parcellation import *
from test_plot import *
from test_treatment import *
from test_glm import *
from test_parallel import *
from test_jde_multi_subj import *
from test_paradigm import *
from test_rfir import *
from test_xml import *
from test_jde_vem_bold import *
from test_ndarray import *

if __name__ == "__main__":
    unittest.main()
